// mpapi.h - memory police application perogrammer's interface
// Owner:  Nikolaus Seydel, Amadeus Data Processing
// Date:    10 Jul 2009
// Summary: mpolice (memory police) can be used to detect strange memory
//          (heap) usage, eg. leaking heap, functional heap leakage, etc.
//          mpolice comes with an executable (mpolice) and with a shared
//          library (libmpolice).
// See mpolice.c for more details

#ifndef _MPAPI_H_
#define _MPAPI_H_

// struct for passing delta information back to API user
// addr is the (32 or 64 byte) memory address of the current block
// pbtrace is a pointer to the function backtrace, an array of void *,
//   terminated by a NULL
// size is the user size of the current memory block

struct mpapi {
    void   *addr;      // memory chunk address
    void  **pbtrace;   // pointer to backtrace, an array of void*
    long    size;      // memory chunk size
};

#ifdef __cplusplus
extern "C" {
#endif

// API function for starting mpolice
// id: a unique positive value for identifying this start/stop cycle
// successful return: 0
// error returns:
//   if id is 0, or negative, -1 is returned
//   if another start is active, that start's positive id is returned

int mpapistart(int id);

// API function for stopping mpolice and taking a delta snapshot, ie.
// all mpolice owned memory chunks allocated since last delta
// id: a unique positive value for identifying this start/stop cycle
// successful return:
//   a pointer to a constant array of struct mpapi items
//   the last item is empty and has addr==NULL, pbtrace==NULL, and size==0
// error return: NULL (no current start, or current start not for id)
// notes:
//   1. the memory returned is owned by mpolice, do not try to write or free it.
//   2. the memory returned will be reused on next successful stop call, do not
//      keep references to that memory around for longer than that.

const struct mpapi *mpapistop(int id);

// API function for getting symbol information from backtrace data
// pbtrace: a pointer to an array of void* backtrace information
// return: a pointer to constant character array containing readable
//   symbol information per backtrace address, separated by \n
// notes:
//   1. the memory returned is owned by mpolice, do not try to write or free it.
//   2. the memory returned will be reused on next successful symbol call, do not
//      keep references to that memory around for longer than that.
//   3. the symbol calculation is done exactly as for mpolice's delta and status
//      processing, ie. --exclude expressions are honoured, and the backtrace
//      is limited to the --nest value at mpolice startup.

const char *mpapisymbols(void * const *pbtrace);

#ifdef __cplusplus
} // extern "C"
#endif

// Sample code.
// Assume a global static variable int apiid which can change
// for each iteration.

/* start of sample code **

// Sample code for starting mpolice

  int returnedid = mpapistart(apiid);
  if (returnedid!=apiid)
  {
    // process mpapistart error
  }

// Sample code for stopping mpolice

  const struct mpapi *papi;
  void *const        *pbt;
  const char         *psymbols;

  papi = mpapistop(apiid);
  if (papi==NULL)
  {
    // process mpapistop error
    return;
  }

  // loop through memory items as long as item.addr is not NULL

  for ( ; papi->addr; ++papi)
  {
    fprintf(stderr, "%08X %6d:", papi->addr, papi->size);

    // loop through backtrace items until a NULL item is hit
    for (pbt = papi->pbtrace; *pbt; ++pbt)
    {
       fprintf(stderr, " %08X", *pbt);
    } // for pbt
    fprintf(stderr, "\n");

    // print backtrace symbols
    psymbols = mpapisymbols(papi->pbtrace);
    if (psymbols==NULL)
    {
      // process mpapisymbols error
    }
    else fprintf(stderr, "%s\n", psymbols);
  } // for papi

** end of sample code */

#endif // #ifndef _MPAPI_H_
