// mpcmd.c  memory policing - mpolice command processor
// Author:  Nikolaus Seydel, Amadeus Data Processing
// Dates:   Created 19 Jan 2009, last changed 03 Jul 2009
// Summary: mpolice (memory police) can be used to detect strange memory
//          (heap) usage, eg. leaking heap, functional heap leakage, etc.
//          mpolice comes with an executable command processor (mpolice),
//          with a shared library (libmpolice), and with a signal generator
//          for remote control (mpsignal).
// Usage:   1. Start the executable to be policed using the mpolice command:
//             > mpolice [any mpolice options] executable parm1 parm2 ...
//          2. Optionally send signals to created process using mpsignal, eg:
//             > mpsignal --pid=1395 delta
//             in order to remotely request mpolice actions
//          3. After end of process, look to mpolice logfile for output
// Options: See help screen below for information on possible options.
//          mpolice sets environment variables MPOLICE_xxxxx, which can
//          be set manually, as well.
 
#include "mpolice.h"   // mpolice internals
#include <getopt.h>    // for command line parameters
#include <errno.h>     // errno for error messages
 
// usage info
 
static const char *version_line =
   "AMPolice 1.4 (c) Amadeus Data Processing\n"
   "For more info, Please log on to http://pmweb.muc.amadeus.net/AMPolice";
 
static const char *help_screen =
    "AMPolice: amadeus memory utilization tool V 1.4\n"
    "usage:   ampolice [options] exe parm1 parm2 ...\n"
    "long opt   short  description\n"
    "--active    -a    activate from the start, default is inactive\n"
    "--begin=n   -b n  begin policing only after n allocations\n"
    "--bytes     -B    output byte counts, not human readable\n"
    "--clear     -c    clear all options read so far\n"
    "--check=i   -C i  check for commands every so many allocations\n"
    "--debug=s   -d s  debug string, see mplib.c for details\n"
    "--disptrac  -D    display trace in log, default is inactive\n"
    "--exclude=e -x e  expression you want to avoid logs\n"
    "--every=i   -e i  log delta or status every i seconds\n"
    "--file=f    -f f  logfile path\n"
    "--high=n    -h n  allocations must be <=b to be policed\n"
    "--help      -H    display help screen and exit\n"
    "--low=n     -l n  allocations must be >=b to be policed\n"
    "--libdir=d  -L d  directory to find libmpolice.so\n"
    "--nest=i    -n i  backtrace nesting depth\n"
    "--procfs=d  -p d  proc filesystem mount point (or 'no')\n"
    "--parent    -P    stay with parent on fork() calls\n"
    "--quit=i    -q i  quit policing after i minutes\n"
    "--random    -r    random sampling (+-percentage)\n"
    "--read=f    -R f  read options from argument file\n"
    "--sample=n  -s n  sample only once per n allocations\n"
    "--status    -S    use every for status, not delta output\n"
    "--symbols=s -y s  display function backtrace symbols on:\n"
    "                  x=exit, e=every, c=command, eg. -y xc\n"
    "--threads   -t    run mpolice in a thread-safe way\n"
    "--try       -T    try and check setup and exit\n"
    "--verbose   -v    run in verbose mode\n"
    "--version   -V    display mpolice version info and exit\n"
    "--write=f   -W f  write current options to argument file\n"
    "n=number: int[k|m], eg. 10, 077, 0x30, 50k, 21m, 0x10m\n"
    "i: int, d: directory, f: file path, s: string of characters\n"
;
 
// option control
 
static const char *short_opts = "ab:BcC:d:De:f:h:Hl:L:n:p:Pq:rR:s:StTvVW:x:y:";
static struct option long_opts[] = {
    { "active",   no_argument,       NULL, 'a' },
    { "begin",    required_argument, NULL, 'b' },
    { "bytes",    no_argument,       NULL, 'B' },
    { "clear",    no_argument,       NULL, 'c' },
    { "check",    required_argument, NULL, 'C' },  // V1.5
    { "debug",    required_argument, NULL, 'd' },
    { "disptrac", no_argument,       NULL, 'D' },  // change Sagar
    { "exclude",  required_argument, NULL, 'x' },  // change Sagar
    { "every",    required_argument, NULL, 'e' },
    { "file",     required_argument, NULL, 'f' },
    { "high",     required_argument, NULL, 'h' },
    { "help",     no_argument,       NULL, 'H' },
    { "low",      required_argument, NULL, 'l' },
    { "libdir",   required_argument, NULL, 'L' },
    { "nest",     required_argument, NULL, 'n' },
    { "procfs",   required_argument, NULL, 'p' },
    { "parent",   no_argument,       NULL, 'P' },
    { "quit",     required_argument, NULL, 'q' },
    { "random",   no_argument,       NULL, 'r' },
    { "read",     required_argument, NULL, 'R' },
    { "sample",   required_argument, NULL, 's' },
    { "status",   no_argument,       NULL, 'S' },
    { "symbols",  required_argument, NULL, 'y' },
    { "threads",  no_argument,       NULL, 't' },
    { "try",      no_argument,       NULL, 'T' },
    { "verbose",  no_argument,       NULL, 'v' },
    { "version",  no_argument,       NULL, 'V' },
    { "write",    required_argument, NULL, 'W' },
    { NULL, 0, NULL, 0 }
};
 
// static variables to hold command options
static int  active_start;        // -a --active  active policing
static int  begin_count;        // -b --begin   begin policing
static int  machine_format;     // -B --bytes   output byte counts
static int  poll_count;         // -C --check   polling countdown
static char debug_string[20];   // -d --debug   debug string
static int  display_trace;        // -D --disptrac display backtrace
static int  every_seconds;      // -e --every   log seconds
static char exclude_string[256];// -x --exclude exclude string
static char log_file_name[256]; // -f --file    log file (stderr, if="")
static char lib_directory[256]; // -l --libdir  libmpolice.so directory
static int  max_chunk_size;     // -h --high    high end of policing
static int  min_chunk_size;     // -l --low     low end of policing
static int  nesting_levels;     // -n --nest    backtrace nesting depth
static char procfs_mount[256];  // -p --procfs  procfs mount point
static int  fork_parent;        // -P --parent  fork: stay with parent
static int  quit_minutes;       // -q --quit    quit minutes
static int  random_sampling;    // -r --random  random sampling
static int  sampling_interval;  // -s --sample  sample interval
static int  default_status;     // -S --status  every=status output
static int  thread_safe;        // -t --threads run thread-safe
static int  verbose_mode;       // -v --verbose run verbose mode
static char show_symbols[4];    // -y --symbols backtrace symbols
 
// other static variables
 
static int  env_ok;             // environment setup worked
 
// forward declarations
 
static int  parse_options(int argc, char *argv[]);
static int  prepare_environment(void);
static void reset_options(void);
static int  execute_option(int readok, int opt, const char *arg);
 
// entry point function
 
int main(int argc, char *argv[])
{
    int          ok, argp;
    const char  *pc = "POSIXLY_CORRECT", *p;
    char        *pcv = NULL;
 
    // save POSIXLY_CORRECT setting
 
    p = getenv(pc);
    if (p) {
        pcv = malloc(strlen(p)+1);
        if (pcv==NULL) exit(16);
        strcpy(pcv, p);
    }
    setenv(pc, "1", 1);
 
    // parse options
 
    reset_options();
    argp = parse_options(argc, argv);
    if (argp<0) return 1;
 
    // undo POSIXLY correct setting
 
    if (pcv) setenv(pc, pcv, 1);
    else unsetenv(pc);
 
    // set environment variables
 
    ok = prepare_environment();
    if (argv[argp]==NULL) return 0;
    if (ok<0) return 3;
 
    // start executable
 
    ok = execvp(argv[argp], argv+argp);
    if (ok==-1) {
        perror("execv failed");
        return 4;
    }
 
    return ok; // should never get here
}
 
// environment functions
 
// set MPOLICE_id from input string val
// return 0 on success, -1 on error
 
static int set_env_string(const char *id, const char *val)
{
    int  ok;
    char name[32];
 
    sprintf(name, "MPOLICE_%s", id);
    ok = setenv(name, val, 1);
    if (ok<0) {
        env_ok = 0;
        fprintf(stderr, "error %d trying to export %s=%s\n",
                errno, name, val);
    }
 
    if (ok>=0 && verbose_mode!=0) {
        fprintf(stderr, "export %s=%s\n", name, val);
    }
 
    return ok;
}
 
// set MPOLICE_id from input integer val
// return 0 on success, -1 on error
 
static int set_env_int(const char *id, int val)
{
    char intval[40];
 
    sprintf(intval, "%d", val);
    return set_env_string(id, intval);
}
 
// set MPOLICE_id from input boolean val
// return 0 on success, -1 on error
 
static int set_env_bool(const char *id, int val)
{
    return set_env_string(id, val==0 ? "NO" : "YES");
}
 
// prepare environment for exec* function call
// return 0 on success, -1 on error
 
static int prepare_environment(void)
{
    char        preload[1024];
    const char *mp = "libmpolice.so";
    const char *oldpreload;
 
    set_env_bool   ("ACTIVE",  active_start);
    set_env_int    ("BEGIN",   begin_count);
    set_env_bool   ("BYTES",   machine_format);
    set_env_int    ("CHECK",   poll_count);
    set_env_string ("DEBUG",   debug_string);
    set_env_bool   ("DISPTRAC", display_trace);
    set_env_string ("EXCLUDE", exclude_string);
    set_env_int    ("EVERY",   every_seconds);
    set_env_string ("FILE",    log_file_name);
    set_env_int    ("HIGH",    max_chunk_size);
    set_env_int    ("LOW",     min_chunk_size);
    set_env_int    ("NEST",    nesting_levels);
    set_env_string ("PROCFS",  procfs_mount);
    set_env_int    ("QUIT",    quit_minutes);
    set_env_bool   ("RANDOM",  random_sampling);
    set_env_int    ("SAMPLE",  sampling_interval);
    set_env_bool   ("STATUS",  default_status);
    set_env_bool   ("THREADS", thread_safe);
    set_env_bool   ("VERBOSE", verbose_mode);
    set_env_string ("SYMBOLS", show_symbols);
    set_env_bool   ("PARENT",  fork_parent);
 
    if (*lib_directory) {
        realpath(lib_directory, preload);
        strcat(preload, "/");
    }
    else {
        preload[0] = '\0';
    }
    strcat(preload, mp);
 
    oldpreload = getenv("LD_PRELOAD");
    if (oldpreload) {
        strcat(preload, " ");
        strcat(preload, oldpreload);
    }
 
    setenv("LD_PRELOAD", preload, 1);
    if (verbose_mode) fprintf(stderr, "export LD_PRELOAD=%s\n", preload);
 
    return env_ok;
}
 
// parsing functions
 
// get MPOLICE_name's value
// return pointer to value, or "", if not defined
 
static const char *env2str(const char *name)
{
    const char *p;
    char envname[80];
 
    sprintf(envname, "MPOLICE_%s", name);
    p = getenv(envname);
    return p==NULL ? "" : p;
}
 
// get MPOLICE_name's value
// return integer value, or def, if not defined
 
static int env2int(const char *name, int def)
{
    const char *p = env2str(name);
    return *p ? atoi(p) : def;
}
 
// get MPOLICE_name's value
// return boolean value (0 or 1), or def, if not defined
 
static int env2bool(const char *name, int def)
{
    char         val[4];
    const char  *p = env2str(name);
    int          i, n = strlen(p);
 
    if (n<2 || n>3) return def;
    strcpy(val, p);
    for (i=0; i<n; ++i) val[i] = tolower(val[i]);
    if (!strcmp(p, "yes")) return 1;
    if (!strcmp(p, "no")) return 0;
    return def;
}
 
// reset all options
 
static void reset_options(void)
{
    active_start      = 0;
    begin_count       = 0;
    poll_count        = 0;
    machine_format    = 0;
    debug_string[0]   = '\0';
    display_trace     = 0;
    exclude_string[0] = '\0';
    every_seconds     = 0;
    log_file_name[0]  = '\0';
    max_chunk_size    = DEF_HIGH;
    min_chunk_size    = DEF_LOW;
    lib_directory[0]  = '\0';
    nesting_levels    = DEF_NEST;
    procfs_mount[0]   = '\0';
    quit_minutes      = 0;
    random_sampling   = 0;
    sampling_interval = 0;
    default_status    = 0;
    thread_safe       = 0;
    verbose_mode      = 0;
    show_symbols[0]   = '\0';
    fork_parent       = 0;
}
 
// read all MPOLICE_xxx environment variables
// copy strings to static storage (for later editing)
 
static void read_environment(void)
{
    const char *pDebug, *pFile, *pLib, *pProc, *pSym,*pExclude; //change Sagar
 
    active_start      = env2int ("ACTIVE",0);
    begin_count       = env2int ("BEGIN", 0);
    machine_format    = env2bool("BYTES", 0);
    poll_count        = env2int ("CHECK", 0);
    pDebug            = env2str ("DEBUG");
    display_trace      = env2int ("DISPTRAC",0);
    every_seconds     = env2int ("EVERY", 0);
    pFile             = env2str ("FILE");
    max_chunk_size    = env2int ("HIGH", DEF_HIGH);
    min_chunk_size    = env2int ("LOW", DEF_LOW);
    pLib              = env2str ("LIBDIR");
    nesting_levels    = env2int ("NEST", DEF_NEST);
    pProc             = env2str ("PROCFS");
    quit_minutes      = env2int ("QUIT", 0);
    random_sampling   = env2bool("RANDOM", 0);
    sampling_interval = env2int ("SAMPLE", 0);
    default_status    = env2bool("STATUS", 0);
    thread_safe       = env2bool("THREADS", 0);
    verbose_mode      = env2bool("VERBOSE", 0);
    pSym              = env2str ("SYMBOLS");
    fork_parent       = env2bool("PARENT", 0);
    pExclude          = env2str ("EXCLUDE");
 
    if (strlen(pDebug)    <sizeof(debug_string) )  strcpy(debug_string,    pDebug);
    if (strlen(pFile)     <sizeof(log_file_name))  strcpy(log_file_name,   pFile);
    if (strlen(pLib)      <sizeof(lib_directory))  strcpy(lib_directory,   pLib);
    if (strlen(pProc)     <sizeof(procfs_mount) )  strcpy(procfs_mount,    pProc);
    if (strlen(pSym)      <sizeof(show_symbols) )  strcpy(show_symbols,    pSym);
    if (strlen(pExclude)  <sizeof(exclude_string)) strcpy(exclude_string,  pExclude);
    if (nesting_levels>MAX_NEST) nesting_levels = MAX_NEST;
}
 
// get a string from an input string
// dsc = description for input (used for possible error message)
// v = input string containing value
// x = pointer to receiving string
// max_len = maximum acceptable input string length
// return length of string (or -1 on error)
 
static int get_string(const char *dsc, const char *v, char *x, int max_len)
{
    // get string argument
 
    int n = strlen(v);
    if (n>max_len-1) {
        fprintf(stderr, "%s input too long: %s\n", dsc, v);
        return -1;
    }
 
    strcpy(x, v);
    return n;
}
 
// get an integer value from an input string
// dsc = description for input (used for possible error message)
// v = input string containing value
// accept_unit = accept an optional K (kilo) or M (mega) unit
// return integer value (or -1 on error)
 
static int get_int(const char *dsc, const char *v, int kfactor)
{
    // get integer argument
    // examples:
    // 100   --> 100 (decimal, no unit)
    // 077   --> 63 (octal, o unit)
    // 0x10  --> 16 (hexadecimal, no unit)
    // 30k   --> 30720 (decimal, 'k'=1024 unit)
    // 013K  --> 11264 (octal, 'K'=1024 unit)
    // 0x10M --> 16777216 (hex, 'M'=1024*1024 unit)
    // kfactor: k means times 1000, 1024, or 0 (not allowed)
 
    int64_t      bignum;
    long         num, factor=0;
    char        *pnext, ch;
 
    num = strtol(v, &pnext, 0);
    if (pnext==v) {
        fprintf(stderr, "%s invalid integer %s\n", dsc, v);
        return -1;
    }
 
    ch = *pnext;
    if (ch=='\0') factor = 1;
    else if (kfactor>0) {
        if (ch=='k' || ch=='K') factor = kfactor;
        else if (ch=='m' || ch=='M') factor = kfactor * kfactor;
    }
 
    if (factor==0) {
        fprintf(stderr, "%s invalid unit %c in %s\n", dsc, ch, v);
        return -1;
    }
 
    if (factor>1 && *(pnext+1)!='\0') {
        fprintf(stderr, "%s invalid char after unit in %s\n", dsc, v);
        return -1;
    }
 
    bignum = num;
    bignum *= factor;
    if (bignum<0 || bignum>INT32_MAX) {
        fprintf(stderr, "%s number %s out of range\n", dsc, v);
        return -1;
    }
    return (int) bignum;
}
 
// display input argument, and exit process
 
static void show_and_exit(const char *p)
{
    printf("%s", p);
    exit(1);
}
 
// try to work out heap chunk size and check some sizeof's
// display success or failure and exit process
 
static void try_and_check(void)
{
    int32_t  errors = 0, sz, bits;
    void    *p;
 
    // check platform
 
    if (!strcmp(PLATFORM, "Unknown")) {
        ++errors;
        printf("Unknown platform\n");
    }
 
    // check void * size
 
    sz = sizeof(void *);
    if (sz==4) bits = 32;
    else if (sz==8) bits = 64;
    else {
        ++errors;
        printf("sizeof(void *) = %d, expected: 4 or 8\n", sz);
    }
 
    // check size_t's size
 
    sz = sizeof(size_t);
    if (sz!=4 && sz!=8) {
        ++errors;
        printf("sizeof(size_t) = %d, expected: 4 or 8\n", sz);
    }
 
    // check heap chunk prefix's size location
 
    p = __libc_malloc(800);
    sz = MSIZE(p);
    if (sz<800 || sz>1024) {
        ++errors;
        printf("heap chunk size (800) not locatable, found %d\n", sz);
    }
 
    // checks done, show summary
 
    if (errors==0) printf("mpolice try+check worked on %d bit %s system\n"
                          "go ahead and use libmpolice\n", bits, PLATFORM);
    else printf("%d check%s failed - cannot use libmpolice\n",
                errors, errors==1 ? "" : "s");
 
    exit(1);
}
 
// read a file with option lines plus optional comments:
//   boolean option, eg. VERBOSE    # comment
//   option=value, eg.   LOW=64#comment
//   comments, eg.       # just a comment
//   empty lines
//   white space is acceptable at all ends, eg. __LOW_=\t_64___
// return 0 (success) or -1
 
static int read_file(const char *path)
{
    int   ok;
    char  buffer[1024], *option, *arg, *p;
    const struct option *plongopt;
    FILE *f = fopen(path, "r");
 
    if (f==NULL) {
        fprintf(stderr, "cannot open read file %s\n", path);
        return -1;
    }
 
    if (verbose_mode) fprintf(stderr, "about to read option file %s\n", path);
    while (!feof(f)) {
        if (NULL==fgets(buffer, sizeof(buffer), f)) break;
 
        // discard trailing comment
        p = strchr(buffer, '#');
        if (p) *p = '\0';
 
        // discard leading and trailing whitespace
        if (buffer[0]=='\0') continue;
        for (p=buffer+strlen(buffer)-1; isspace(*p); --p) *p = '\0';
        for (p=buffer; *p && isspace(*p); ++p) *p = '\0';
        if (*p=='\0') continue;
 
        // isolate option and argument
        ok = 1;
        option = p;
        p = strchr(p, '=');
        if (p==NULL) arg = NULL;
        else {
            arg = p+1;
            while (isspace(*(p-1))) --p;
            *p = '\0';
            while (isspace(*arg)) ++arg;
        }
 
        // translate option to lowercase
        for (p=option; *p; ++p) *p = tolower(*p);
 
        // find option
        for (plongopt=long_opts; plongopt->name!=NULL; ++plongopt) {
            if (!strcmp(plongopt->name, option)) break;
        }
        if (plongopt->name==NULL) {
            fprintf(stderr, "%s: invalid option %s\n",
                    path, option);
            ok = 0;
        }
        if (ok) {
            if (plongopt->has_arg==required_argument && arg==NULL) {
                fprintf(stderr, "%s: missing argument on %s\n",
                        path, option);
                ok = 0;
            }
            if (plongopt->has_arg==no_argument && arg!=NULL) {
                fprintf(stderr, "%s: invalid argument on %s\n",
                        path, option);
                ok = 0;
            }
        }
        if (ok) execute_option(0, plongopt->val, arg);
    } // while (!feof(f))
 
    fclose(f);
    if (verbose_mode) fprintf(stderr, "done reading option file %s\n", path);
    return 0;
}
 
// write a file with option lines, eg. VERBOSE, or LOW=64
// return 0 (success) or -1
 
static int write_file(const char *path)
{
    time_t     now = time(0);
    struct tm *t   = gmtime(&now);
    FILE      *f   = fopen(path, "w");
 
    if (f==NULL) {
        fprintf(stderr, "cannot open write file %s\n", path);
        return -1;
    }
 
    fprintf(f, "# %04d-%02d-%02d %02d.%02d.%02d %s",
            t->tm_year+1900, t->tm_mon+1, t->tm_mday,
            t->tm_hour, t->tm_min, t->tm_sec, version_line);
 
    // boolean variables are first
 
    if (active_start)    fprintf(f, "active\n");
    if (display_trace)   fprintf(f, "disptrac\n");
    if (machine_format)  fprintf(f, "bytes\n");
    if (random_sampling) fprintf(f, "random\n");
    if (default_status)  fprintf(f, "status\n");
    if (thread_safe)     fprintf(f, "threads\n");
    if (verbose_mode)    fprintf(f, "verbose\n");
    if (fork_parent)     fprintf(f, "parent\n");
 
    // integer variables are next
 
    if (begin_count>0)            fprintf(f, "begin=%d\n", begin_count);
    if (poll_count>0)             fprintf(f, "check=%d\n", poll_count);
    if (every_seconds>0)          fprintf(f, "every=%d\n", every_seconds);
    if (max_chunk_size!=DEF_HIGH) fprintf(f, "high=%d\n", max_chunk_size);
    if (min_chunk_size!=DEF_LOW)  fprintf(f, "low=%d\n", min_chunk_size);
    if (nesting_levels!=DEF_NEST) fprintf(f, "nest=%d\n", nesting_levels);
    if (quit_minutes>0)           fprintf(f, "quit=%d\n", quit_minutes);
    if (sampling_interval>0)      fprintf(f, "sample=%d\n", sampling_interval);
 
    // string variables are last
 
    if (*debug_string)  fprintf(f, "debug=%s\n",   debug_string);
    if (*log_file_name) fprintf(f, "file=%s\n",    log_file_name);
    if (*lib_directory) fprintf(f, "libdir=%s\n",  lib_directory);
    if (*procfs_mount)  fprintf(f, "procfs=%s\n",  procfs_mount);
    if (*show_symbols)  fprintf(f, "symbols=%s\n", show_symbols);
    if (*exclude_string)fprintf(f, "exclude=%s\n", exclude_string);
 
    // that was it
 
    fclose(f);
    if (verbose_mode) fprintf(stderr, "options written to %s\n", path);
    return 0;
}
 
// execute an option
// read_ok = --read is an acceptable option
// opt = short option character to execute
// arg = possible argument, maybe NULL
// return 0 (success) or -1
 
static int execute_option(int read_ok, int opt, const char *arg)
{
    int ok, len;
 
    switch (opt) {
    case '?':  // getopt_long error
        break;
    case 'b':  // -b --begin
        begin_count = get_int("--begin:", arg, 1000);
        if (begin_count<0) begin_count = 0;
        break;
    case 'B':  // -B --bytes
        machine_format = 1;
        break;
    case 'c':  // -c --clear
        reset_options();
        break;
    case 'C':  // -C --check
        poll_count = get_int("--check:", arg, 0);
        if (poll_count<0) poll_count = 0;
        break;
    case 'd':  // -d --debug
        ok = get_string("--debug:", arg, debug_string,
                        sizeof(debug_string));
        if (ok<0) strcpy(debug_string, "");
        break;
    case 'e':  // -e --every
        every_seconds = get_int("--every:", arg, 1000);
        if (every_seconds<0) every_seconds = 0;
        break;
    case 'f':  // -f --file
        ok = get_string("--file:", arg, log_file_name,
                        sizeof(log_file_name));
        if (ok<0) *log_file_name = '\0';
        break;
    case 'h':  // -h --high
        max_chunk_size = get_int("--high:", arg, 1024);
        if (max_chunk_size<0) max_chunk_size = DEF_HIGH;
        break;
    case 'H':  // -H --help
        show_and_exit(help_screen);
        break;
    case 'l':  // -l --low
        min_chunk_size = get_int("--low:", arg, 1024);
        if (min_chunk_size<0) min_chunk_size = DEF_LOW;
        break;
    case 'L':  // -L --library
        ok = get_string("--library:", arg, lib_directory,
                        sizeof(lib_directory));
        if (ok<0) *lib_directory = '\0';
        break;
    case 'n':  // -n --nest
        nesting_levels = get_int("--nest:", arg, 0);
        if (nesting_levels<0) nesting_levels = DEF_NEST;
        if (nesting_levels>MAX_NEST) nesting_levels = MAX_NEST;
        break;
    case 'p':  // -p --procfs
        ok = get_string("--procfs:", arg, procfs_mount,
                        sizeof(log_file_name));
        if (ok<0) *procfs_mount = '\0';
        len = strlen(procfs_mount);
        if (len>1 && procfs_mount[len-1]=='/') procfs_mount[len-1] = '\0';
        if (!strcmp(procfs_mount, "/proc")) *procfs_mount = '\0';
        break;
    case 'P':  // -P --parent
        fork_parent = 1;
        break;
    case 'q':  // -q --quit
        quit_minutes = get_int("--quit:", arg, 0);
        if (quit_minutes<0) quit_minutes = 0;
        break;
    case 'r':  // -r --random
        random_sampling = 1;
        break;
    case 'R':  // -R --read
        if (!read_ok) fprintf(stderr, "nested --read is not allowed");
        else read_file(arg);
        break;
    case 's':  // -s --sample
        sampling_interval = get_int("--sample:", arg, 1000);
        if (sampling_interval<0) sampling_interval = 0;
        break;
    case 'S':  // -S --status
        default_status = 1;
           break;
       case 'a':
           active_start=1; // -a -active
           break;
       case 'D':
           display_trace=1;
    case 't':  // -t --threads
        thread_safe = 1;
        break;
    case 'T':  // -T --try
        try_and_check();
        break;
    case 'v':  // -v --verbose
        verbose_mode = 1;
        break;
    case 'V':  // -V --version
        show_and_exit(version_line);
        break;
    case 'W':  // -W --write
        write_file(arg);
        break;
    case 'x':  // -W --write   //Sagar adding
         ok = get_string("--exclude:", arg, exclude_string,
                                 sizeof(exclude_string));
        if (ok<0) strcpy(exclude_string, "");
        break;
    case 'y':  // -y --symbols
        ok = get_string("--symbols:", arg, show_symbols,
                        sizeof(show_symbols));
        if (ok<0) strcpy(show_symbols, "");
        break;
    default:
        fprintf(stderr, "unexpected getopt_long error\n");
    }
 
    return 0;
}
 
// parse options from command line input arguments
// return index of first positional argument, or -1 (error)
// if --help/-h or --version/-V was found, execute
// option and exit process as required by GNU standards
 
static int parse_options(int argc, char *argv[])
{
    int opt, index;
 
    read_environment();
    while (1) {
        opt = getopt_long(argc, argv, short_opts, long_opts, &index);
        if (opt<0) return optind;
        if (execute_option(1, opt, optarg)<0) return -1;
    }
 
    return -1; // we should never get here
}
