// mplib.c  memory policing - shared library
// Owner:  Nikolaus Seydel, Amadeus Data Processing
// Date:    19 Jan 2009
// Changed: 14 Jul 2009
// Summary: mpolice (memory police) can be used to detect strange memory
//          (heap) usage, eg. leaking heap, functional heap leakage, etc.
//          mpolice comes with an executable (mpolice) and with a shared
//          library (libmpolice).
// See mpolice.c for more details
// Debugging characters:
//   C=Chain    - checks mpolice owned chunk chain on entry and exit to
//                allocation functions (malloc/calloc/realloc/free)
//   R=Realloc  - checks different aspects of realloc (no/no, no/mp, etc.)
//   L=Libc     - shows libc function addresses at initialization time
//   F=Function - print one line per intercepted function call

static const char *version_line =
    "libmpolice.so.1.5 (c) Amadeus Data Processing Jan 2009 \n";

#include "mpolice.h"
#include "mpapi.h"    //ADDED V1.5 Sagar Mark2
#include <pwd.h>
#include <sys/stat.h>
#include <sched.h>
#include <regex.h>   //ADDED V1.2 Sagar Mark1

#if USE_SIGNALS==1
#include <signal.h>
#endif

#include <execinfo.h>
#include <assert.h>
#include <pthread.h>

#define VP(ptr) voidp2str(ptr)

// some definitions

#define PFX_ID "AMP"        // Amadeus Memory Police
#define PENDING_DELTA 0x01  // pending DELTA
#define PENDING_STAT  0x02  // pending STATUS
#define PENDING_CMD   0x04  // pending remote command
#define PROCESS_EVERY 0x01  // processing EVERY
#define PROCESS_CMD   0x02  // processing COMMAND
#define PROCESS_EXIT  0x04  // processing EXIT
#define UNCHECKED     '*'   // unchecked mpolice chunk
#define CHECKED       'C'   // checked mpolice chunk

// mpolice's heap chunk prefix

struct mp_prefix {
    struct mp_prefix *next;  // next chunk in chain
    struct mp_prefix *prev;  // previous chunk in chain
    void *backtrace[2];      // backtrace
};

// note that backtrace size is actually a run time constant nesting_levels
// After nesting_levels backtrace items, there is a int32 (initial size)
// followed by a char[4] with an identifier to indicate mp_prefix:
// mpolice's heap chunk prefix, part 2 - Size And ID

struct mp_said {
    uint32_t usize;          // user requested size
    char     chk;            // checked at last delta?
    char     id[3];          // identifier
};

#define PFXSIZE ((2+nesting_levels)*sizeof(void *)+sizeof(struct mp_said))

// static variables read from environment variables

static int32_t active_start         = 0;  // active start for logging
static int32_t display_trace     = 0;  // display trace while logging
static int32_t output_bytes      = 0;  // output byte counts
static int32_t check_every       = 0;  // command file check count
static int32_t poll_countdown    = 0;  // command file poll countdown
static int32_t thread_safe       = 0;  // run thread-safe
static int32_t verbose_mode      = 0;  // run in verbose mode
static int32_t default_status    = 0;  // run output_status as default
static int32_t show_symbols_x    = 0;  // show bt symbols at exit
static int32_t show_symbols_c    = 0;  // show bt symbols at command
static int32_t show_symbols_e    = 0;  // show bt symbols at every
static int32_t fork_parent       = 0;  // stay with parent on fork()

static int32_t min_size          = 0;  // low end of policing
static int32_t max_size          = 0;  // high end of policing
static int32_t begin_cnt         = 0;  // begin after
static int32_t quit_mins         = 0;  // quit minutes
static int32_t every_secs        = 0;  // log seconds
static int32_t nesting_levels    = 0;  // backtrace nesting
static int32_t random_incr       = 0;  // random sampling
static int32_t sampling_interval = 0;  // log sampling
static int32_t iRegular_Check    = 0;  // if expression can be compared or not Sagar
static int32_t iString_Check     = 0;  // string check enabled or not

static const char *logfile_path = "";  // log file path
static const char *debug_string = "";  // debug string
static const char *proc_mount   = "";  // proc fs mount point
static const char *show_symbols = "";  // show bt symbols
static const char *exclude_string="";  // exclude string
static regex_t     compiled_expression; //exclude string Sagar

// static variables for internal use

static int32_t candidate_cnt     = 0;  // candidate allocations
static int32_t next_to_log       = 0;  // next to log
static int32_t malloc_cnt        = 0;  // total malloc count
static int32_t calloc_cnt        = 0;  // total calloc count
static int32_t realloc_cnt       = 0;  // total realloc count
static int32_t free_cnt          = 0;  // total free count
static int32_t last_malloc_cnt   = 0;  // last delta malloc count
static int32_t last_calloc_cnt   = 0;  // last delta calloc count
static int32_t last_realloc_cnt  = 0;  // last delta realloc count
static int32_t last_free_cnt     = 0;  // last delta free count
static int32_t chain_cnt         = 0;  // number of chained chunks
static int32_t pending_mask      = 0;  // pending output
static int32_t processing_ind    = 0;  // processing indicator
static int32_t new_item_cnt      = 0;  // new items count

// static API data V1.5 Start Sagar Mark1

static int     api_current_start = 0;  // current api start id
static int     api_delta_size    = 0;  // API delta buffer size
static int     api_btrace_size   = 0;  // API backtrace buffer size
static int     api_symbol_size= 65535; // API symbol buffer size
static struct mpapi
              *api_delta_buf   = NULL; // API delta buffer
static void  **api_btrace_buf  = NULL; // API backtrace buffer
static char   *api_symbol_buf  = NULL; // API symbols buffer

// static API data V1.5 END Sagar Mark1
// miscellaneous static storage

static unsigned long first_address;   // first function address
static unsigned long last_address;    // last function address

static long    heap_at_delta     = 0;  // mpolice heap at last delta
static long    old_chunk_bytes   = 0;  // delta: old chunk bytes
static long    new_chunk_bytes   = 0;  // delta: new chunk bytes

static __thread int32_t internal_depth = 0;  // each thread's depth
static int32_t big_endian        = 0;  // big (or little) endian
static time_t  next_out_time     = 0;  // next status/delta output
static time_t  quit_time         = 0;  // time to quit
static time_t  delta_time        = 0;  // last delta output
static time_t  current_time      = 0;  // current time
static char    signal_file[256]  = ""; // signal file
static FILE   *log_file        = NULL; // log file
static pid_t   my_pid            = 0;  // my process ID

// list head for mpolice owned chunks

static struct mp_prefix list_head = { &list_head, &list_head };

// static debug variables

static const char *func_name     = ""; // last function name
static int32_t debug_print       = 0;  // print debugging info
static int32_t debug_chain       = 0;  // C - chain debugging
static int32_t debug_realloc     = 0;  // R - realloc debugging
static int32_t debug_libc        = 0;  // L - libc debugging
static int32_t debug_func        = 0;  // F - function debugging
static int32_t reanull_cnt       = 0;  // calls to reallocate(NULL, ...)
static int32_t reanono_cnt       = 0;  // calls to reanono
static int32_t reampno_cnt       = 0;  // calls to reampno
static int32_t reanomp_cnt       = 0;  // calls to reanomp
static int32_t reampmp_cnt       = 0;  // calls to reampmp

// thread-safe storage

static pthread_mutex_t mpolice_mutex = PTHREAD_MUTEX_INITIALIZER;

// thread-safe routines

static void get_mutex(void)
{
    if (internal_depth==0 && thread_safe) pthread_mutex_lock(&mpolice_mutex);
    ++internal_depth;
}

static void release_mutex(void)
{
    --internal_depth;
    if (internal_depth==0 && thread_safe) pthread_mutex_unlock(&mpolice_mutex);
}

#if USE_SIGNALS==1
// signal handler

__mpextern void handle_signal(int signo)
{
    if (signo==SIG_DEFAULT) {
        if (default_status) pending_mask |= PENDING_STAT;
        else pending_mask |= PENDING_DELTA;
    }
    else if (signo==SIG_COMMAND) pending_mask |= PENDING_CMD;
}
#endif

// prefix movement functions

static struct mp_said *pfx2said(struct mp_prefix *ppfx)
{
    return (struct mp_said *) (ppfx->backtrace+nesting_levels);
}

static void *said2user(struct mp_said *psaid)
{
    return (void *) (psaid->id + sizeof(psaid->id));
}

static struct mp_prefix *user2pfx(void *puser)
{
    char *pchar = (char *) puser;
    return (struct mp_prefix *) (pchar-PFXSIZE);
}

// common helper functions

static void check_chain(void)
{
    int               i;
    struct mp_prefix *pthis;
    struct mp_said   *psaid;

    for (i=0, pthis=&list_head; i<chain_cnt; ++i) {
        pthis = pthis->next;
        psaid = pfx2said(pthis);
        assert(psaid->chk==UNCHECKED || psaid->chk==CHECKED);
        assert(memcmp(psaid->id, PFX_ID, sizeof(psaid->id))==0);
    }
    assert(pthis->next==&list_head);

    for (i=0, pthis=&list_head; i<chain_cnt; ++i) {
        pthis = pthis->prev;
    }
    assert(pthis->prev==&list_head);
}

static void check_on_entry(void)
{
    check_chain();
}

static void check_on_exit(void)
{
    check_chain();
}

static const char *now(void)
{
    static char dt_now[24];
    struct tm  *t = localtime(&current_time);

    sprintf(dt_now, "%04d-%02d-%02d %02d.%02d.%02d",
            t->tm_year+1900, t->tm_mon+1, t->tm_mday,
            t->tm_hour, t->tm_min, t->tm_sec);

    return dt_now;
}

static const char *getenv_ex(const char *env_name)
{
    const char *p = getenv(env_name);
    if (p==NULL) p = "";
    return p;
}

static int do_log(int32_t sz)
{
    // return whether logging is appropriate

    if (internal_depth!=1) return 0;
    if (sz<min_size || sz>max_size) return 0;

    ++candidate_cnt;
    if (candidate_cnt<begin_cnt || current_time>quit_time) return 0;
    if (sampling_interval) {
        int delta;

        if (candidate_cnt<next_to_log) return 0;
        delta = (int) sampling_interval;
        if (random_incr) {
             int plusminus = rand() % random_incr;
             if ((rand() % 2)==0) delta += plusminus;
             else delta -= plusminus;
        }
        next_to_log += delta;
    }
    return 1;
}

// output functions

static const char *voidp2str(void *p)
{
    static int      index = 0;
    static char     data[80];
    char           *out;

    out = data + index;
    index += 20;
    if (index>sizeof(data)-20) index = 0;
    sprintf(out, PTRFMT, (long) p);
    return out;
}

static void print_libc(const char *which, void *fmp, void *flc)
{
    fprintf(stderr, "libmpolice::%-7s at 0x%s, __libc_%-7s at 0x%s\n",
            which, voidp2str(fmp), which, voidp2str(flc));
}

static int is_big_endian(void)
{
    int    n;
    char  *p;

    n = 1;
    p = (char *) &n;
    return p[0]==1 ? 0 : 1;
}

static void print_void_ptr(void *p)
{
    fprintf(log_file, " %s", voidp2str(p));
}

static void print_human(long sz)
{
    double d;

    if (output_bytes) {
        fprintf(log_file, " %12ld", sz);
        return;
    }

    if (sz<1024) fprintf(log_file, " %6ld", sz);
    else if (sz<1000*1024) {
        d = (1.0 * sz) / 1024.0;
        fprintf(log_file, " %5.1fK", d);
    }
    else if (sz<10000*1024) {
        d = (1.0 * sz) / 1024.0;
        fprintf(log_file, " %5.0fK", d);
    }
    else if (sz<1000*1024*1024) {
        d = (1.0 * sz) / (1024.0 * 1024.0);
        fprintf(log_file, " %5.1fM", d);
    }
    else {
        d = (1.0 * sz) / (1024.0 * 1024.0);
        fprintf(log_file, " %5.0fM", d);
    }
}

static void output_strparm(const char *title, const char *value)
{
    fprintf(log_file, "%-26s : %s\n", title, value);
}

static void output_boolparm(const char *title, int val)
{
    output_strparm(title, val!=0 ? "YES" : "NO");
}

static void output_parameters(void)
{
    const struct passwd *pw;
    uid_t uid, eid;
    char  value[80];

    fprintf(log_file, "%s mpolice started pid-%d\n", now(), my_pid);
    fprintf(log_file, "%s", version_line);
    sprintf(value, "%s %d-bit %s endian",
            PLATFORM, __WORDSIZE, big_endian ? "big" : "little");
    output_strparm("mpolice platform", value);

    uid = getuid();
    eid = geteuid();
    if (uid!=eid) {
        pw = getpwuid(uid);
        sprintf(value, "%d %s", uid, pw ? pw->pw_name : "<unknown>");
        output_strparm("real user", value);
        pw = getpwuid(eid);
        sprintf(value, "%d %s", eid, pw ? pw->pw_name : "<unknown>");
        output_strparm("effective user", value);
    }
    else {
        pw = getpwuid(uid);
        sprintf(value, "%d %s", uid, pw ? pw->pw_name : "<unknown>");
        output_strparm("real and effective user", value);
    }

    output_strparm("fork() behaviour",
        fork_parent!=0 ? "stay with parent" : "follow child");

    sprintf(value, "%d .. %d", min_size, max_size);
    output_strparm("track heap chunks of size", value);
    sprintf(value, "%d", begin_cnt);
    output_strparm("begin tracking after", value);
    sprintf(value, "%d", nesting_levels);
    output_strparm("backtrace nesting levels", value);

    if (check_every>0) {
      sprintf(value, "%d", check_every);
      output_strparm("check cmd file every", value);
    }

    if (sampling_interval>1) {
        sprintf(value, "%d%s", sampling_interval,
            random_incr>0 ? " (random)" : "");
        output_strparm("sampling interval", value);
    }
    else output_strparm("sampling", "OFF");

    if (every_secs>0) {
        sprintf(value, "%d sec (%s)",
            every_secs, default_status!=0 ? "status" : "delta");
        output_strparm("automatic output every", value);
    }
    else output_strparm("automatic (every) output", "OFF");

    if (quit_mins>0) {
        sprintf(value, "%d min", quit_mins);
        output_strparm("quit after", value);
    }

    output_boolparm("mpolice active from start", active_start);
    output_boolparm("mpolice will log traces", display_trace);
    output_boolparm("thread-safe implementation", thread_safe);
    output_strparm("output backtrace symbols", show_symbols);

    if(strlen(exclude_string)>0)
    {
        iString_Check=1;
        output_strparm("exclude string ",exclude_string);
        iRegular_Check=regcomp(&compiled_expression, exclude_string, REG_EXTENDED);
        if (iRegular_Check!=0) {
            output_strparm("string check disabled", exclude_string);
        }
        else
        {
            output_strparm("string check enabled", exclude_string);
        }
    }
    else
    {
        iString_Check=0;
        output_strparm("string check enabled", "NO");
    }

    fprintf(log_file, "%s end of mpolice parameters\n", now());
}

static void output_heap_status(void)
{
    fprintf(log_file, "function calls: "
            "%d malloc, %d calloc, %d realloc, %d free\n",
            malloc_cnt, calloc_cnt, realloc_cnt, free_cnt);
}

static void output_symbols(void * const *bt)
{
    int          i, size;
    char       **symbol;

    for (size=0; size<nesting_levels; ++size) if (bt[size]==NULL) break;
    if (size==0) return;
    //CHN_ADD_SAGAR_START
    if(size==1)
    {
        fprintf(log_file, "\n");
    }
    else
    {
        symbol = backtrace_symbols(bt, size);
        for (i=0; i<size; ++i)
        {
            if(iString_Check==1)
            {
                if(iRegular_Check==0)
                {
                    if(0!= regexec(&compiled_expression, symbol[i], 0, NULL, 0))
                    {
                        fprintf(log_file, "        %s\n", symbol[i]);
                    }
                }
            }
            else
            {
                fprintf(log_file, "        %s\n", symbol[i]);
            }

        }
    }
    //CHN_ADD_SAGAR_END


    __libc_free(symbol);
    ++free_cnt;
}

static void output_chain(int all)
{
    int32_t           i, j, shown=0, suppressed=0, showsym=0;
    struct mp_prefix *pthis = &list_head;
    struct mp_said   *psaid;

    fprintf(log_file, "   i. c %sblksze %s address%s\n",
            output_bytes ? "      " : "",
            sizeof(void*)==4 ? "" : "        ",
            nesting_levels==0 ? "" : " backtrace");

    if (processing_ind==PROCESS_EXIT  && show_symbols_x) showsym = 1;
    if (processing_ind==PROCESS_CMD   && show_symbols_c) showsym = 1;
    if (processing_ind==PROCESS_EVERY && show_symbols_e) showsym = 1;

    old_chunk_bytes  = 0;
    new_chunk_bytes  = 0;
    new_item_cnt = 0;
    for (i=0; i<chain_cnt; ++i) {
        pthis = pthis->next;
        psaid = pfx2said(pthis);
        if (all || psaid->chk==UNCHECKED) {
            ++new_item_cnt;
            new_chunk_bytes += psaid->usize;
            if (shown<MAX_OUTPUT) {
                fprintf(log_file, "%4d. %c", new_item_cnt, psaid->chk);
                print_human((long) psaid->usize);
                print_void_ptr(said2user(psaid));
                for (j=0; j<nesting_levels; ++j)
                    print_void_ptr(pthis->backtrace[j]);
                fprintf(log_file, "\n");
                if (showsym) output_symbols(pthis->backtrace);
                ++shown;
            }
            else ++suppressed;
            if (!all) psaid->chk = CHECKED;
        }
        else old_chunk_bytes += psaid->usize;
    }

    if (suppressed>0) fprintf(log_file,
        "*** more than %d items, %d item%s suppressed\n",
        MAX_OUTPUT, suppressed, suppressed==1 ? "" : "s");
}

static void output_line(const char *desc, long n, const char *unit)
{
    char ch;

    if (n<0) {
        ch = '-';
        n = -n;
    }
    else ch = ' ';

    fprintf(log_file, "%-26s : %c", desc, ch);
    print_human(n);
    fprintf(log_file, " %s\n", unit);
}

static void switch_logfile(const char *file)
{
    static FILE *save_logfile;

    if (file) {
        FILE *f = fopen(file, "w");
        if (f==NULL) {
            fprintf(stderr, "mpolice cannot open %s for command output\n", file);
            return;
        }
        save_logfile = log_file;
        log_file = f;
    }
    else {
        if (log_file!=save_logfile) fclose(log_file);
        log_file = save_logfile;
    }
}

static void output_status(const char *title, const char *file)
{
    if (file) switch_logfile(file);
    fprintf(log_file, "%s mpolice status (%s)\n", now(), title);
    output_heap_status();
    fprintf(log_file, "%d mpolice controlled heap chunks:\n", chain_cnt);
    if (chain_cnt) output_chain(1);
    fprintf(log_file, "mpolice controlled heap bytes:");
    print_human(old_chunk_bytes + new_chunk_bytes);
    fprintf(log_file, "\n");
    fprintf(log_file, "%s end of mpolice status.\n", now());
    fflush(log_file);
    if (file) switch_logfile(NULL);
    pending_mask &= ~PENDING_STAT;
    return;
}

static void output_delta(const char *title, const char *file)
{
    long heap, incr;
    int  secs, mallocs, callocs, reallocs, frees, missing;
    char descr[80];

    time(&current_time);
    secs = current_time - delta_time;
    if (secs<=0) secs = 1;

    if (file) switch_logfile(file);
    fprintf(log_file, "%s mpolice delta (%s)\n", now(), title);
    if (display_trace) output_chain(0);
    heap = old_chunk_bytes + new_chunk_bytes;
    incr = heap - heap_at_delta;
    mallocs  = malloc_cnt  - last_malloc_cnt;
    callocs  = calloc_cnt  - last_calloc_cnt;
    reallocs = realloc_cnt - last_realloc_cnt;
    frees    = free_cnt    - last_free_cnt;
    missing  = mallocs + callocs - frees;

    fprintf(log_file, "function calls since last delta:\n");
    fprintf(log_file, "%d malloc, %d calloc, %d realloc, %d free\n",
            mallocs, callocs, reallocs, frees);
    output_line("time since last delta", secs, "seconds");
    if (missing>1) output_line("missing frees", missing, "calls");
    output_line("last delta mpolice heap", heap_at_delta, "bytes");
    output_line("current mpolice heap", heap, "bytes");
    sprintf(descr, "%d new chunks", new_item_cnt);
    output_line(descr, new_chunk_bytes, "bytes");
    output_line("increment since then", incr, "bytes");
    output_line("relative increment", incr/secs, "bytes per second");

    fprintf(log_file, "%s end of mpolice delta.\n", now());
    fflush(log_file);
    if (file) switch_logfile(NULL);
    pending_mask &= ~PENDING_DELTA;
    delta_time = current_time;
    heap_at_delta = heap;
    last_malloc_cnt = malloc_cnt;
    last_calloc_cnt = calloc_cnt;
    last_realloc_cnt = realloc_cnt;
    last_free_cnt = free_cnt;
    return;
}

static void output_settings(void)
{
    char sample_string[80], plusminus_string[10], quit_string[80];
    const char *p_sample, *p_quit;
    struct tm *t;

    if (quit_time==0) p_quit = "*stopped*";
    else if (quit_time==INT32_MAX) p_quit = "*never*";
    else {
        t = gmtime(&quit_time);
        sprintf(quit_string, "%02d:%02d:%02d",
                t->tm_hour, t->tm_min, t->tm_sec);
        p_quit = quit_string;
    }

    if (sampling_interval==1) p_sample = "*no sampling*";
    else {
        sprintf(sample_string, "%10d", sampling_interval);
        if (random_incr>0) {
            sprintf(plusminus_string, " +-%d%%", RND_SAMPLE);
            strcat(sample_string, plusminus_string);
        }
        p_sample = sample_string;
    }

    fprintf(stderr, "%s mpolice settings\n", now());
    fprintf(stderr,
            "low    - %10d\n"
            "high   - %10d\n"
            "sample - %s\n"
            "quit   - %s\n",
            min_size, max_size, p_sample, p_quit);
}

static void output_proc_maps(void)
{
    char maps_file[80], line[1025];
    FILE *f;

    if (!strcmp(proc_mount, "no")) {
        fprintf(log_file, "%s maps: not copied (--procfs=no)\n", now());
        return;
    }

    sprintf(maps_file, "%s/self/maps", *proc_mount ? proc_mount : "/proc");
    f = fopen(maps_file, "r");
    if (f==NULL) {
        fprintf(log_file, "%s maps: cannot open %s\n", now(), maps_file);
        return;
    }

    fprintf(log_file, "%s maps: copy %s\n", now(), maps_file);
    while (!feof(f)) {
        if (NULL==fgets(line, sizeof(line), f)) break;
        fputs(line, log_file);
    }
    fclose(f);
    fprintf(log_file, "%s maps: done %s\n", now(), maps_file);
    fflush(log_file);
}

// API functions V1.5

// API function for starting mpolice
// id: a unique positive value for identifying this start/stop cycle
// successful return: returns input id
// error returns:
//   if id is 0, or negative, -1 is returned
//   if another start is active, that start's positive id is returned

int mpapistart(int id)
{
    if (id<=0) return -1;
    if (api_current_start>0) return api_current_start;
    quit_time = INT32_MAX;
    api_current_start = id;
    return id;
}

// API function for stopping mpolice and taking a delta snapshot, ie.
// all mpolice owned memory chunks allocated since last delta
// id: a unique positive value for identifying this start/stop cycle
// successful return:
//   a pointer to a constant array of struct mpapi items
//   the last item is empty and has addr==NULL, pbtrace==NULL, and size==0
// error return: NULL (no current start, or current start not for id)
// notes:
//   1. the memory returned is owned by mpolice, do not try to write or free it.
//   2. the memory returned will be reused on next successful stop call, do not
//      keep references to that memory around for longer than that.

const struct mpapi *mpapistop(int id)
{
    int32_t           i, j, n;
    struct mp_prefix *pthis = &list_head;
    struct mp_said   *psaid;
    struct mpapi     *pChunk;
    void            **pBacktrace;

    if (api_current_start==0)
    {
		fprintf(stderr, "Nothing to Report : No start for ID ( %d) found \n",id);
		return NULL;
	}


    if(api_current_start!=id)
    {
		fprintf(stderr, "Nothing to Report: Currently ampolice is logging for ID(%d) \n",api_current_start);
		return NULL;
	}
    if (api_delta_buf==NULL) {
        api_delta_size = (MAX_OUTPUT+1) * sizeof(struct mpapi);
        api_delta_buf = __libc_malloc(api_delta_size);
        if (api_delta_buf==NULL)
        {
			fprintf(stderr, "ampolice:ampapistop failed : malloc error");
			return NULL;
		}
        api_btrace_size = MAX_OUTPUT * (nesting_levels+1) * sizeof(void *);
        api_btrace_buf = __libc_malloc(api_btrace_size);
        if (api_btrace_buf==NULL)
        {
			fprintf(stderr, "ampolice:ampapistop failed : malloc error");
			return NULL;
		}
    }

    pChunk = api_delta_buf;
    pBacktrace = api_btrace_buf;
    n = 0;

    for (i=0; i<chain_cnt; ++i) {
        pthis = pthis->next;
        psaid = pfx2said(pthis);
        if (n<MAX_OUTPUT && psaid->chk==UNCHECKED) {
            pChunk->addr = said2user(psaid);
            pChunk->pbtrace = pBacktrace;
            pChunk->size = psaid->usize;

            for (j=0; j<nesting_levels; ++j)
               pBacktrace[j] = pthis->backtrace[j];

            pBacktrace[nesting_levels] = NULL;
            pBacktrace += nesting_levels+1;
            ++n;
            ++pChunk;
        }
        psaid->chk = CHECKED;
    }

    pChunk->addr = NULL;
    pChunk->pbtrace = NULL;
    pChunk->size = 0;

    api_current_start = 0;
    quit_time = (time_t) 0;
    return api_delta_buf;
}

// API function for getting symbol information from backtrace data
// pbtrace: a pointer to an array of void* backtrace information
// return: a pointer to constant character array containing readable
//   symbol information per backtrace address, separated by \n
// notes:
//   1. the memory returned is owned by mpolice, do not try to write or free it.
//   2. the memory returned will be reused on next successful symbol call, do not
//      keep references to that memory around for longer than that.
//   3. the symbol calculation is done exactly as for mpolice's delta and status
//      processing, ie. --exclude expressions are honoured, and the backtrace
//      is limited to the --nest value at mpolice startup.

const char *mpapisymbols(void * const *pbtrace)
{
    int            i, out, len, size, left;
    void * const  *pbt;
    char         **symbol;
    char          *pout;

    if (api_symbol_buf==NULL) {
        api_symbol_buf = __libc_malloc(api_symbol_size);
        if (api_symbol_buf==NULL)
        {
			fprintf(stderr, "ampolice:ampapisymbols failed : malloc error");
			return NULL;
		}
    }
    pout = api_symbol_buf;
    left = api_symbol_size;

    for (size=0, pbt=pbtrace; *pbt!=NULL; ++pbt) ++size;
    symbol = backtrace_symbols(pbtrace, size);
       for (i=0; i<size; ++i) {
        if (iString_Check==1 && iRegular_Check==0)
            out = (0!= regexec(&compiled_expression, symbol[i], 0, NULL, 0));
        else out = 1;

        if (out) {
            len = strlen(symbol[i]);
            if (len>=left) break;
            strcpy(pout, symbol[i]);
            pout += len;
            left -= len;
        }
    }

    __libc_free(symbol);
    return api_symbol_buf;
}

// run remote command

static void run_command()
{
    int   i, n, exit_proc;
    FILE *f;
    char  line[1024], imsg[80], *p, *cmd, *arg;
    const char *msg = imsg;
    time_t tnow, tmax;

    pending_mask &= ~PENDING_CMD;
    processing_ind = PROCESS_CMD;
    exit_proc = 0;

    f = fopen(signal_file, "r");

    if (f==NULL) {
        fprintf(stderr, "cannot open %s\n", signal_file);
        return;
    }

    fgets(line, sizeof(line), f);
    fclose(f);
    n = strlen(line);
    if (n<=2) return;

    // rewrite signal file for future use

    tnow = time(0);
    tmax = tnow + RETRY_SECS;
    while (tnow<=tmax) {
        for (i=0; i<10; ++i) {
              f = fopen(signal_file, "w");
            if (f) break;
        }
        if (f) break;
        sched_yield();
        tnow = time(0);
    }

    if (f==NULL) {
        fprintf(stderr, "cannot rewrite %s\n", signal_file);
        return;
    }
    fprintf(f, "\n");
    fclose(f);

    // normalize command

    if (line[n-1]=='\n') line[n-1] = '\0';
    cmd = line;
    if (!isalpha(cmd[0])) return;

    p = strchr(line, ' ');
    if (p==NULL) arg = "";
    else { *p = '\0'; arg = p+1; }

    if (!strcmp("id", cmd)) {
        fprintf(log_file, "%s ID: %s\n", now(), arg);
        fflush(log_file);
        fprintf(stderr, "%s ID line written to log file\n", now());
        return;
    }

    fprintf(log_file, "%s remote command: %s %s\n", now(), cmd, arg);
    if (!strcmp(cmd, "quit")) {
        quit_time = (time_t) 0;
        msg = "stopped policing";
    }
    if (!strcmp(cmd, "default")) cmd = default_status ? "status" : "delta";

    if (!strcmp(cmd, "tracestart")) {
        display_trace = 1;
        msg = "Now you will get the trace/symbol information";
    }
    else if (!strcmp(cmd, "tracestop")) {
        display_trace = 0;
        msg = "Now you will not get the trace/symbol information";
    }
    else if (!strcmp(cmd, "start")) {
        quit_time = INT32_MAX;
        msg = "started policing";
    }
    else if (!strcmp(cmd, "exit")) {
        exit_proc = 1;
        msg = "will exit now";
    }
    else if (!strcmp(cmd, "query")) {
        output_settings();
        msg = "settings have been displayed";
    }
    else if (!strcmp(cmd, "delta")) {
        if (*arg=='\0') arg = NULL;
        output_delta("command", arg);
        msg = "logged current delta";
    }
    else if (!strcmp(cmd, "status")) {
        if (*arg=='\0') arg = NULL;
        output_status("command", arg);
        msg = "logged current status";
    }
    else if (!strcmp(cmd, "low")) {
        min_size = atoi(arg);
        sprintf(imsg, "set LOW to %d", min_size);
    }
    else if (!strcmp(cmd, "high")) {
        max_size = atoi(arg);
        sprintf(imsg, "set HIGH to %d", max_size);
    }
    else if (!strcmp(cmd, "sample")) {
        sampling_interval = atoi(arg);
        if (sampling_interval<=0) sampling_interval = 1;
        if (random_incr && sampling_interval!=1) {
            random_incr = (sampling_interval * RND_SAMPLE / 100);
            if (random_incr<2) random_incr = 2;
        }
        else random_incr = 0;
        sprintf(imsg, "set SAMPLE to %d", sampling_interval);
    }
    else sprintf(imsg, "encountered invalid cmd %s", cmd);

    fprintf(stderr, "%s mpolice %s\n", now(), msg);
    processing_ind = 0;
    if (exit_proc) {
        --internal_depth;
        exit(12);
    }
}

static void run_remote_cmd()
{
   ++internal_depth;
   run_command();
   --internal_depth;
}

// timer and pending commands

static void check_time(void)
{
    time_t       now;
    struct stat  statbuf;
    int          rc;

    if (internal_depth!=1) return;//Fix 1.4 Sagar

    // get current time
    // check file, if polling countdown reached 0
    // check file, if last check was not at current time

    time(&now);
    if (check_every>0) {
      --poll_countdown;
      if (poll_countdown<=0) {
          current_time = (time_t) 0;
          poll_countdown = check_every;
      }
    }
    if (now==current_time) return;

    // check for signal file changes

    current_time = now;
    rc = stat(signal_file, &statbuf);
    if (rc==0 && statbuf.st_size>2) pending_mask |= PENDING_CMD;

    // run output_status or output_delta, if the time is due

    if (every_secs==0) return;
    if (current_time<next_out_time) return;
    processing_ind = PROCESS_EVERY;
    next_out_time = current_time + every_secs;
    if (default_status) output_status("every", NULL);
    else output_delta("every", NULL);
    processing_ind = 0;
}

static void process_pending(void)
{
    if (pending_mask & PENDING_STAT)  output_status("signal", NULL);
    if (pending_mask & PENDING_DELTA) output_delta("signal", NULL);
    if (pending_mask & PENDING_CMD)   run_remote_cmd();
}

// cleanup at end of process, must be before allocation functions

static void __finisattr mpolice_fini(void)
{
    get_mutex();
    ++internal_depth;

    if (my_pid==getpid()) {
        process_pending();
        if (debug_realloc) fprintf(log_file,
            "realloc usage: %d null, %d nono, %d mpno, %d nomp, %d mpmp\n",
            reanull_cnt, reanono_cnt, reampno_cnt, reanomp_cnt, reampmp_cnt);
        processing_ind = PROCESS_EXIT;
        output_status("finish", NULL);
        processing_ind = 0;
        remove(signal_file);
        //output_proc_maps();
        fprintf(log_file, "%s mpolice ended.\n", now());
        fflush(log_file);
        if (*logfile_path!='\0') fclose(log_file);
        if (api_delta_buf!=NULL) __libc_free(api_delta_buf);
        if (api_btrace_buf!=NULL) __libc_free(api_btrace_buf);
        if (api_symbol_buf!=NULL) __libc_free(api_symbol_buf);
    }

    --internal_depth;
    release_mutex();
    FINI_MUTEX(mpolice_mutex)
}

// initialize heap chunk prefix

static void *init_prefix(struct mp_prefix *pthis, int32_t sz)
{
    struct mp_said *psaid = pfx2said(pthis);

    if (nesting_levels>0) {
        int32_t        i, j, n;
        unsigned long  fptr;
        void          *bt[MAX_NEST+10];

        n = backtrace(bt, sizeof(bt)/sizeof(bt[0]));
        for (i=0, j=0; i<n && j<nesting_levels; ++i) {
            fptr = (unsigned long) bt[i];
            if (fptr<first_address || fptr>last_address)
                pthis->backtrace[j++] = (void *) fptr;
        }
        for (i=j; i<nesting_levels; ++i) pthis->backtrace[i] = NULL;
    }

    psaid->usize = sz;
    psaid->chk = UNCHECKED;
    memcpy(psaid->id, PFX_ID, sizeof(psaid->id));
    return (void *) (psaid->id + sizeof(psaid->id));
}

// realloc specific functions

static void *reanomp(void *ptr, size_t size)
{
    // ptr not mpolice owned, new block mpolice owned

    void             *p;
    char             *pchar;
    struct mp_prefix *pthis, *pprev, *pnext;

    if (debug_realloc) { func_name = "reanomp"; ++reanomp_cnt; }
    p = __libc_realloc(ptr, size+PFXSIZE);
    if (p==NULL) return p;

    // shift user data to user portion of new block

    pchar = (char *) p;
    memmove(pchar+PFXSIZE, pchar, size);

    // add new block to mpolice chain

    pthis = (struct mp_prefix *) p;
    pprev = list_head.prev;
    pnext = &list_head;
    pthis->next = pnext;
    pthis->prev = pprev;
    pprev->next = pthis;
    pnext->prev = pthis;
    ++chain_cnt;

    // initialize mpolice prefix

    p = init_prefix(pthis, size);
    return p;
}

static void *reampmp(struct mp_prefix *pmp, size_t size)
{
    // ptr mpolice owned, new block mpolice owned

    void             *p;
    struct mp_prefix *pthis, *pprev, *pnext;
    struct mp_said   *psaid;

    if (debug_realloc) { func_name = "reampmp"; ++reampmp_cnt; }
    p = __libc_realloc(pmp, size+PFXSIZE);
    if (p==NULL) return p;

    pthis = (struct mp_prefix *) p;
    psaid = pfx2said(pthis);

    // set to unchecked, new size, prev's next, and next's prev

    pprev = pthis->prev;
    pnext = pthis->next;

    psaid->chk = UNCHECKED;
    psaid->usize = size;
    pprev->next = pthis;
    pnext->prev = pthis;

    return said2user(psaid);
}

static void *reampno(struct mp_prefix *pmp, size_t size)
{
    // ptr mpolice owned, new block not mpolice owned

    void             *p;
    struct mp_prefix *pthis, *pprev, *pnext;
    struct mp_said   *psaid;

    if (debug_realloc) { func_name = "reampno"; ++reampno_cnt; }
    p = __libc_realloc(pmp, size+PFXSIZE);
    if (p==NULL) return p;

    // remove old block from mpolice chain

    pthis = (struct mp_prefix *) p;
    pprev = pthis->prev;
    pnext = pthis->next;
    pprev->next = pnext;
    pnext->prev = pprev;
    --chain_cnt;

    // shift user data to newly allocated block

    psaid = pfx2said(pthis);
    memmove(pthis, said2user(psaid), size);

    return p;
}

// internal allocation functions

static void *mp_malloc(size_t sz)
{
    // allocate sz bytes plus mpolice prefix
    // fill in mpolice prefix
    // note that the mutex must be held when calling mp_malloc

    void *p;
    struct mp_prefix *pthis, *pprev, *pnext;

    p = __libc_malloc(sz+PFXSIZE);
    if (p==NULL) return p;

    pthis = (struct mp_prefix *) p;
    pprev = list_head.prev;
    pnext = &list_head;

    pthis->next = pnext;
    pthis->prev = pprev;
    pprev->next = pthis;
    pnext->prev = pthis;
    ++chain_cnt;

    p = init_prefix(pthis, sz);
    return p;
}

static size_t mp_free(void *ptr)
{
    // free an mpolice owned or a standalone heap chunk
    // remove from mpolice chain, if owned by mpolice
    // note that the mutex must be held when calling mp_free

    size_t sz;
    void *p;
    struct mp_prefix *pthis, *pprev, *pnext;
    struct mp_said *psaid;

    if (ptr==NULL) return 0;

    pthis = user2pfx(ptr);
    psaid = (struct mp_said *) (pthis->backtrace+nesting_levels);
    if (0!=memcmp(psaid->id, PFX_ID, sizeof(psaid->id))) {
        sz = MSIZE(ptr);
        p = ptr;
    }
    else {
        ++free_cnt;
        if (debug_func) fprintf(stderr, "mpolice: free(0x%s)\n", VP(ptr));
        sz = psaid->usize;
        pprev = pthis->prev;
        pnext = pthis->next;
        pprev->next = pnext;
        pnext->prev = pprev;
        --chain_cnt;
        p = (void *) pthis;
    }

    __libc_free(p);
    return sz;
}

static void *mp_realloc(void *ptr, size_t size, size_t *oldsz)
{
    int               ownit;
    void             *p;
    char             *pchar;
    struct mp_prefix *pthis;
    struct mp_said   *psaid;

    debug_print = 0;

    // case 1: ptr==NULL, ie. realloc is like malloc

    if (ptr==NULL) {
        if (debug_realloc) { func_name = "reanull"; ++reanull_cnt; }
        *oldsz = 0;
        ownit = do_log(size);
        debug_print = debug_func && ownit;
        if (ownit) p = mp_malloc(size);
        else p = __libc_malloc(size);
        return p;
    }

    pchar = (char *) ptr;
    pthis = (struct mp_prefix *) (pchar - PFXSIZE);
    psaid = (struct mp_said *) (pthis->backtrace+nesting_levels);

    // case 2: ptr is not mpolice controlled

    if (0!=memcmp(psaid->id, PFX_ID, sizeof(psaid->id))) {
        *oldsz = MSIZE(ptr);
        ownit = do_log(size);
        debug_print = debug_func && ownit;
        if (ownit) p = reanomp(ptr, size);
        else {
            p = __libc_realloc(ptr, size);
            if (debug_realloc) { func_name = "reanono"; ++reanono_cnt; }
        }
        return p;
    }

    // case 3: ptr is mpolice controlled

    debug_print = debug_func;
    *oldsz = psaid->usize;
    if (do_log(size)) p = reampmp(pthis, size);
    else p = reampno(pthis, size);
    return p;
}

// intercepted functions

void *cALLOc(size_t nmemb, size_t size)
{
    size_t  sz = nmemb * size;
    int     ownit;
    void   *p;

    get_mutex();
    if (debug_chain) check_on_entry();

    ownit = do_log(sz);
    if (ownit) {
        ++calloc_cnt;
        p = mp_malloc(sz);
        if (p) memset(p, '\0', sz);
    }
    else p = __libc_calloc(nmemb, size);

    check_time();
    if (pending_mask) process_pending();
    if (debug_chain) check_on_exit();

    if (debug_func && ownit) fprintf(stderr,
        "mpolice: calloc(%d, %d) --> 0x%s\n",
        (int32_t) nmemb, (int32_t) size, VP(p));

    release_mutex();
    return p;
}

void *mALLOc(size_t size)
{
    int     ownit;
    void   *p;

    get_mutex();
    if (debug_chain) check_on_entry();

    ownit = do_log(size);
    if (ownit)
    {
        p = mp_malloc(size);
        ++malloc_cnt;
    }
    else p = __libc_malloc(size);

    check_time();
    if (pending_mask) process_pending();
    if (debug_chain) check_on_exit();

    if (debug_func && ownit) fprintf(stderr,
        "mpolice: malloc(%d) --> 0x%s\n", (int32_t) size, VP(p));

    release_mutex();
    return p;
}

void fREe(void *ptr)
{
    size_t sz;

    get_mutex();
    if (debug_chain) check_on_entry();
    //++free_cnt;

    sz = mp_free(ptr);

    check_time();
    if (pending_mask) process_pending();
    if (debug_chain) check_on_exit();
    release_mutex();
}

void *rEALLOc(void *ptr, size_t size)
{
    void   *p;
    size_t  oldsz;

    get_mutex();
    if (debug_chain) check_on_entry();
    ++realloc_cnt;

    p = mp_realloc(ptr, size, &oldsz);

    if (debug_print) fprintf(stderr,
        "mpolice: realloc(0x%s, %d) --> 0x%s\n",
        VP(ptr), (int32_t) size, VP(p));
    check_time();
    if (pending_mask) process_pending();
    if (debug_chain) check_on_exit();

    release_mutex();
    return p;
}

pid_t fORk(void)
{
    pid_t pid;

    get_mutex();
    ++internal_depth;

    time(&current_time);
    pid = __libc_fork();
    if (pid<0) return pid;

    if (pid==0) {

        // child process code

        if (!fork_parent) {
            FILE *f;

            my_pid = getpid();
            sprintf(signal_file, SIG_FILE, my_pid);
            f = fopen(signal_file, "w");
            if (f) { fprintf(f, "\n"); fclose(f); }
            fprintf(log_file, "%s fork: follow child pid-%d\n",
                    now(), my_pid);
            fflush(log_file);
        }
    }

    if (pid>0) {

        // parent process code

        if (fork_parent) {
            fprintf(log_file, "%s fork (pid-%d): stay with parent pid-%d\n",
                    now(), pid, my_pid);
            fflush(log_file);
        }
        else {
            if (*logfile_path) {
                fclose(log_file);
                logfile_path = "";
                log_file = stderr;
            }
            quit_time = 0;
            remove(signal_file);
            my_pid = 0;
        }
    }

    --internal_depth;
    release_mutex();
    return pid;
}

// initialization, must be behind all intercepted functions

__mpextern void __initattr mpolice_init(void)
{
    int              random_sampling;
    unsigned long    func_addr;
    const char      *p;
    FILE            *sig_file;

    // initialize time, random numbers, and big/little endian

    INIT_MUTEX(mpolice_mutex)
    get_mutex();
    ++internal_depth;
    my_pid = getpid();
    time(&current_time);
    srand(current_time);
    big_endian = is_big_endian();

    // initialize signalling support

    sprintf(signal_file, SIG_FILE, my_pid);
    sig_file = fopen(signal_file, "w");
    if (sig_file) {
        fprintf(sig_file, "\n");
        fclose(sig_file);
    }

#if USE_SIGNALS==1
    signal(SIG_DEFAULT, handle_signal);
    signal(SIG_COMMAND, handle_signal);
#endif

    // initialize variables from environment
    active_start      = !strcmp(getenv_ex("MPOLICE_ACTIVE"), "YES");
    display_trace     = !strcmp(getenv_ex("MPOLICE_DISPTRAC"), "YES");
    output_bytes      = !strcmp(getenv_ex("MPOLICE_BYTES"), "YES");
    random_sampling   = !strcmp(getenv_ex("MPOLICE_RANDOM"), "YES");
    thread_safe       = !strcmp(getenv_ex("MPOLICE_THREADS"), "YES");
    verbose_mode      = !strcmp(getenv_ex("MPOLICE_VERBOSE"), "YES");
    fork_parent       = !strcmp(getenv_ex("MPOLICE_PARENT"), "YES");
    default_status    = !strcmp(getenv_ex("MPOLICE_STATUS"), "YES");

    debug_string      = getenv_ex("MPOLICE_DEBUG");
    logfile_path      = getenv_ex("MPOLICE_FILE");
    proc_mount        = getenv_ex("MPOLICE_PROCFS");
    show_symbols      = getenv_ex("MPOLICE_SYMBOLS");
    exclude_string    = getenv_ex("MPOLICE_EXCLUDE");

    min_size          = atoi(getenv_ex("MPOLICE_LOW"));
    max_size          = atoi(getenv_ex("MPOLICE_HIGH"));
    begin_cnt         = atoi(getenv_ex("MPOLICE_BEGIN"));
    check_every       = atoi(getenv_ex("MPOLICE_CHECK"));
    quit_mins         = atoi(getenv_ex("MPOLICE_QUIT"));
    every_secs        = atoi(getenv_ex("MPOLICE_EVERY"));
    sampling_interval = atoi(getenv_ex("MPOLICE_SAMPLE"));
    nesting_levels    = atoi(getenv_ex("MPOLICE_NEST"));
    poll_countdown    = check_every;

    // set random_incr, if sampling is desired

    if (sampling_interval==0) sampling_interval = 1;
    if (random_sampling && sampling_interval!=1) {
        random_incr = (sampling_interval * RND_SAMPLE / 100);
        if (random_incr<2) random_incr = 2;
    }
    else random_incr = 0;

    // set times for --every and --quit

    if (every_secs>0) next_out_time = current_time + every_secs;
    else next_out_time = INT32_MAX;
    if (active_start) quit_time = (time_t) INT32_MAX;
    else quit_time = (time_t) 0;
    if (quit_mins>0) quit_time = current_time + 60 * quit_mins;

    delta_time = current_time;

    // set log file

    if (*logfile_path=='\0') log_file = stderr;
    else {
        char ptrFileTemp[300];
        sprintf(ptrFileTemp, LOG_NEW_NAME,logfile_path,my_pid);
        logfile_path=ptrFileTemp;
        log_file = fopen(logfile_path, "w");

        if (log_file==NULL) {
            log_file = stderr;
            fprintf(stderr, "mpolice could not open %s\n", logfile_path);
            logfile_path = "";
        }
    }
    output_parameters();
    fflush(log_file);

    // get function address boundaries for backtrace exclusion

    func_addr = (unsigned long) &mpolice_fini;
    first_address = last_address = func_addr;
    func_addr = (unsigned long) &mpolice_init;
    if (func_addr<first_address) first_address = func_addr;
    if (func_addr>last_address) last_address = func_addr;

    // set debugging booleans

    for (p=debug_string; *p; ++p) {
        switch (tolower(*p)) {
        case 'c': debug_chain   = 1; break;
        case 'r': debug_realloc = 1; break;
        case 'l': debug_libc    = 1; break;
        case 'f': debug_func    = 1; break;
        }
    }

    // set show_symbol booleans

    for (p=show_symbols; *p; ++p) {
        if (*p=='x' || *p=='X') show_symbols_x = 1;
        if (*p=='e' || *p=='E') show_symbols_e = 1;
        if (*p=='c' || *p=='C') show_symbols_c = 1;
    }

    // display intercepted and intercepting functions

    if (debug_libc) {
        print_libc("malloc",  & mALLOc,  & __libc_malloc);
        print_libc("calloc",  & cALLOc,  & __libc_calloc);
        print_libc("realloc", & rEALLOc, & __libc_realloc);
        print_libc("free",    & fREe,    & __libc_free);
        print_libc("fork",    & fORk,    & __libc_fork);
        fflush(log_file);
    }

    // done with initialization

    setenv("LD_PRELOAD", "", 1);
    malloc_cnt = 0;
    free_cnt = 0;
    --internal_depth;
    release_mutex();
}
