// mpsig.c  memory policing - signal processor for remote commands
// Author:  Nikolaus Seydel, Amadeus Data Processing
// Date:    28 Jan 2009
// Changed: 03 Jul 2009
// Summary: mpsignal used to send a signal to a running process that is
//          linked with libmpolice. Commands are transported by means of a
//          temporary file.
//          Signals are no longer used. Instead, the temporary file is
//          updated, and file updates are recognized by the target process.
// Usage:   mpsignal [pid] [cmd [parameters]]
// Where:   pid is the target process's process id. If only one mpolice
//          process is running, you do not need to specify a process id.
//          cmd [parameters] can be one of the following:
//          (none)     - run default output (could be delta or status, depends
//                       on whether --status was specified on mpolice command
//          quit       - quit mpolicing (continues to free storage, but does
//                       not log any allocations from now on)
//          exit       - exit process
//          query      - display current settings
//          delta [f]  - output current delta to logfile or to file f
//          status [f] - output current status to logfile or to file f
//          low n      - set lower boundary of logged allocations
//          high n     - set higher boundary of logged allocations
//          sample n   - set current sampling interval
//          id x       - write an ID line
//            start       - start mpolicing
//            tracestart - make mpolice display function backtrace
//            tracestop  - make mpolice NOT display function backtrace


#include "mpolice.h"
#include <sys/stat.h>
#include <errno.h>
#include <sched.h>
#include <signal.h>
#include <dirent.h>

#define ARG_NONE    0    // command must not have an argument
#define ARG_OPT     1    // command may or may not have an argument
#define ARG_MAND    2    // command must have an argument
#define ARG_BAD     3    // bad command
#define MAX_RM      3    // max mpolice files to remove

static const char *help_line = "usage: "
    "mpsignal [pid] [cmd [parms]]\n"
    "  pid - process id of target process\n"
    "  cmd - delta/status/start/quit/query/exit/tracestart/tracestop\n"
    "  cmd parms - low n, high n, sample n, id x, delta/status [file]\n"
;

static const char *version_line =
    "ampsignal 1.5 (c) Amadeus Data Processing 2009\n";

// show a line to terminal and exit process

static void show_line(const char *line)
{
    printf("%s", line);
    exit(1);
}

// check command, and whether a required parameter is given
// return 0 (success) or 8

__mpextern int check_command(const char *cmd, int parms)
{
    int need_parms;

    if (0 /* no argument commands */
       || !strcmp(cmd, "")
       || !strcmp(cmd, "quit")
       || !strcmp(cmd, "exit")
       || !strcmp(cmd, "query")
       || !strcmp(cmd, "start")
       || !strcmp(cmd, "tracestart")  //ADD_SAGAR
       || !strcmp(cmd, "tracestop")   //ADD_SAGAR
    ) need_parms = ARG_NONE;
    else if (0 /* mandatory argument commands */
       || !strcmp(cmd, "low")
       || !strcmp(cmd, "high")
       || !strcmp(cmd, "sample")
       || !strcmp(cmd, "id")
    ) need_parms = ARG_MAND;
    else if (0 /* optional argument commands */
       || !strcmp(cmd, "delta")
       || !strcmp(cmd, "status")
    ) need_parms = ARG_OPT;
    else need_parms = ARG_BAD;

    if (parms<=0 && need_parms==ARG_MAND) {
        fprintf(stderr, "missing argument on %s\n", cmd);
        return 8;
    }

    if (parms>0 && need_parms==ARG_NONE) {
        fprintf(stderr, "invalid argument on %s\n", cmd);
        return 8;
    }

    if (need_parms==ARG_BAD) {
        fprintf(stderr, "unknown command %s\n", cmd);
        return 8;
    }

    return 0;
}

// guess pid, if only one mpolice process is running
// remove up to MAX_RM non0-existing mpolice files
// return pid of only process, if applicable, else -1

__mpextern int guess_pid(void)
{
    int             found, len, pid;
    int             i, remove_cnt, remove_file[MAX_RM];
    DIR            *dirp;
    struct dirent  *ep;
    char            dir[80], file[256], *p;
    const char     *name;

    // set dir to SIG_FILE, eg. /tmp/mpolice.%d

    strcpy(dir, SIG_FILE);
    p = strrchr(dir, '/');
    if (p==NULL || p==dir) return -1;

    // set dir to /tmp\0mpolice., point name to mpolice.

    *p = '\0';
    name = p+1;
    *strchr(p+1, '%') = '\0';
    len = strlen(name);

    // find matching files

    found = 0;
    remove_cnt = 0;
    dirp = opendir(dir);
    if (dirp==NULL) return -1;

    while ((ep=readdir(dirp))!=NULL) {
        if (ep->d_type!=0 && ep->d_type!=DT_REG) continue;
        if (memcmp(ep->d_name, name, len)!=0) continue;
        pid = atoi(ep->d_name + len);
        if (pid==0) continue;

        if (kill(pid, 0)==0) {
            if (found==0) found = pid;
            else found = -1;
        }
        else {
            if (remove_cnt<MAX_RM) remove_file[remove_cnt++] = pid;
        }

        if (found<0 && remove_cnt==MAX_RM) break;
    } // while (ep!=NULL)
    closedir(dirp);

    // remove up to MAX_RM files without processes

    for (i=0; i<remove_cnt; ++i) {
        sprintf(file, SIG_FILE, remove_file[i]);
        remove(file);
    }

    // return the one and only pid that was found, or -1

    return found>0 ? found : -1;
}

#if USE_SIGNALS==1
static const char *get_signal_string(int n, const char *def)
{
    if (n==SIGUSR1) return "SIGUSR1";
    if (n==SIGUSR2) return "SIGUSR2";
    return def;
}
#endif

static int send_command(int pid, const char *cmd, char *argv[], int parms)
{
    time_t       now, max;
    int          i, rc;
    FILE        *f;
    char         file[256];
    struct stat  statbuf;

#if USE_SIGNALS==1
    if (*cmd=='\0') return kill(pid, SIG_DEFAULT);
#endif

    sprintf(file, SIG_FILE, pid);

    // wait for the file to have a size of <=2 (\n by libmpolice)

    now = time(0);
    max = now + RETRY_SECS;
    while (now<=max) {
        rc = stat(file, &statbuf);
        if (rc==0 && statbuf.st_size<=2) break;
        if (rc!=0 && errno==ENOENT) {
            statbuf.st_size = 0;
            break;
        }
           sched_yield();
        now = time(0);
    }
    if (statbuf.st_size>2) {
        fprintf(stderr, "libmpolice pid-%d not executing commands\n", pid);
        exit(4);
    }

    // try opening the file several times, in order to serialize with mpolice
    // main process. Try for a max of RETRY_SECS seconds, then give up.

    now = time(0);
    max = now + RETRY_SECS;
    while (now<=max) {
          f = fopen(file, "w");
           if (f) break;
           sched_yield();
        now = time(0);
    }
    if (f==NULL) return -1;

    fprintf(f, "%s", cmd);
    if (parms>0) for (i=parms; argv[i]; ++i) fprintf(f, " %s", argv[i]);

    fprintf(f, "\n");
    fclose(f);

#if USE_SIGNALS==1
    return kill(pid, SIG_COMMAND);
#else
    return 0;
#endif
}

int __main(int argc, char *argv[])
{
    int         index, pid, rc, parms;
    char        msg[80];
    const char *cmd, *sig;

    if (argc>1 && !strcmp(argv[1], "--help")) show_line(help_line);
    if (argc>1 && !strcmp(argv[1], "--version")) show_line(version_line);

    if (argc>1 && isdigit(*argv[1])) {
        pid = atoi(argv[1]);
        if (pid<=0) {
            fprintf(stderr, "parm indicates pid %d, which is not >0\n", pid);
            return 8;
        }
        index = 2;
    }
    else {
        pid = guess_pid();
        if (pid==-1) {
            fprintf(stderr, "mpsignal cannot guess process id\n");
            return 8;
        }
        index = 1;
    }

    if (argc>index) {
        cmd = argv[index];
        if (argc>index+1) parms = index+1;
        else parms = -1;
    }
    else {
        cmd = "";
        parms = -1;
    }

    rc = check_command(cmd, parms);
    if (rc!=0) return rc;
    if (*cmd=='\0') cmd = "default";
    rc = send_command(pid, cmd, argv, parms);

#if USE_SIGNALS==1
    if (*cmd) sig = get_signal_string(SIG_COMMAND, "SIG_COMMAND");
    else sig = get_signal_string(SIG_DEFAULT, "SIG_DEFAULT");
    sprintf(msg, "%s to pid-%d", sig, pid);
    if (rc==0) printf("mpsignal sent %s\n", msg);
    else fprintf(stderr, "mpsignal failed to send %s\n", msg);
#else
    sig = NULL;
    *msg = '\0';
    if (rc==0) printf("mpsignal requested '%s' to pid-%d\n", cmd, pid);
    else printf("mpsignal failed to send '%s' to pid-%d\n", cmd, pid);
#endif

    return rc==0 ? 0 : 4;
}
